-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

-- Rear Angle Control System (RACS)
-- by wlkmanist, 2023

local M = {}
M.type = "auxilliary"
M.relevantDevice = nil

local abs = math.abs
local max = math.max
local min = math.min

local lastIsEnabled = true
local initialAngle = 0                                          -- default steering angle

-- Variables
local fullAngleSpeed = 22                                       -- m/s, always higher than "zeroAngleSpeed"
local zeroAngleSpeed = 6                                        -- m/s, rear wheels turns the same direction as the front if this > current speed, otherwise wheels turns opposite
local steeringSensetivity = 2                                   -- steering angle multiplier (before speed calculations)
local isEnabled = true                                          -- master switch

function updateGFX(dt)
  local steeringInput = electrics.values.steering_input or 0    -- get steering input from player or AI
  local wheelSpeed = math.abs(electrics.values.wheelspeed or 0) -- get current wheel speed
  
  local targetAngle = steeringInput * steeringSensetivity       -- set right angle
  targetAngle = math.max(math.min(targetAngle, 1), -1)          -- constraining
  
  targetAngle = targetAngle * ((wheelSpeed - zeroAngleSpeed) / (fullAngleSpeed - zeroAngleSpeed))   -- add speed dependency with reversed direction at low speeds
  if wheelSpeed < zeroAngleSpeed and zeroAngleSpeed >= 2 then
    local zeroAngleCorrection = fullAngleSpeed / zeroAngleSpeed - 1                                 -- correcting zero angle, so at zero speed steering angle will maximized
    targetAngle = targetAngle * zeroAngleCorrection * 1.25                                          -- apply correction and additional unnecessary constant
  end
  targetAngle = math.max(math.min(targetAngle, 1), -1)          -- constraining
  
  if wheelSpeed < 0.8 or isEnabled == false then       -- disable system at very low speeds or if it is disabled externally
    targetAngle = initialAngle
  end

  if isEnabled ~= lastIsEnabled then
    if isEnabled == true then
      guihooks.message("Rear Angle Control System Enabled", 10, "vehicle.racs.enabled")
    else
      guihooks.message("Rear Angle Control System Disabled", 10, "vehicle.racs.disabled")
    end
    
    lastIsEnabled = isEnabled
  end

  electrics.values.racs = targetAngle                           -- output value
end

local function init(jbeamData)
  if jbeamData.fullAngleSpeed ~= nil then
    fullAngleSpeed = jbeamData.fullAngleSpeed
  end
  if jbeamData.zeroAngleSpeed ~= nil then
    zeroAngleSpeed = jbeamData.zeroAngleSpeed
  end
  if jbeamData.steeringSensetivity ~= nil then
    steeringSensetivity = jbeamData.steeringSensetivity
  end
  if jbeamData.isEnabled ~= nil then
    isEnabled = jbeamData.isEnabled
  end
  
  lastIsEnabled = isEnabled
  electrics.values.racs = initialAngle
end

local function reset(jbeamData)
  electrics.values.racs = initialAngle
end

local function setParameters(parameters)
  if parameters.fullAngleSpeed ~= nil then
    fullAngleSpeed = parameters.fullAngleSpeed
  end
  if parameters.zeroAngleSpeed ~= nil then
    zeroAngleSpeed = parameters.zeroAngleSpeed
  end
  if parameters.steeringSensetivity ~= nil then
    steeringSensetivity = parameters.steeringSensetivity
  end
  if parameters.isEnabled ~= nil then
    isEnabled = parameters.isEnabled
  end
end

M.init = init
M.reset = reset
M.updateGFX = updateGFX
M.setParameters = setParameters

return M
